
<xsl:stylesheet version="1.0" xmlns="http://www.hyfinity.com/xfactory" xmlns:xstore="http://www.hyfinity.com/xstore" xmlns:xg="http://www.hyfinity.com/xgate" xmlns:hy="http://www.hyfinity.com/xengine" xmlns:http="http://schemas.xmlsoap.org/wsdl/http/" xmlns:mime="http://schemas.xmlsoap.org/wsdl/mime/" xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/" xmlns:soap12="http://schemas.xmlsoap.org/wsdl/soap12/" xmlns:soapenc="http://schemas.xmlsoap.org/soap/encoding/" xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/" xmlns:xfact="http://www.hyfinity.com/xfactory" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
    <xsl:output encoding="UTF-8" indent="yes" method="xml" version="1.0"/>
    <!--	match on root template to begin processing-->
    <xsl:variable name="Chosen_Name_Prefix">
        <xsl:value-of select="/xg:message/xg:header/xg:attributes/xg:attribute[@xg:name='Chosen_Name_Prefix']"/>
    </xsl:variable>
    <xsl:variable name="Chosen_Project_Layer">
        <xsl:value-of select="/xg:message/xg:header/xg:attributes/xg:attribute[@xg:name='Chosen_Project_Layer']"/>
    </xsl:variable>
    <xsl:variable name="Chosen_Collection">
        <xsl:value-of select="/xg:message/xg:header/xg:attributes/xg:attribute[@xg:name='Chosen_Collection']"/>
    </xsl:variable>
    <xsl:variable name="Chosen_Pattern_Instance_Pool">
        <xsl:value-of select="/xg:message/xg:header/xg:attributes/xg:attribute[@xg:name='Chosen_Pattern_Instance_Pool']"/>
    </xsl:variable>
    <xsl:template match="/">
        <Control_Pattern_Entries xmlns="http://www.hyfinity.com/xfactory">
            <xsl:apply-templates mode="control_pattern_addition" select="/xg:message/xg:body/hy:XStore_Response/xstore:xstore/xstore:content/wsdl:definitions/wsdl:service"/>
        </Control_Pattern_Entries>
    </xsl:template>
    <!-- The following script generates the correct data into the blueprint required for Service integration pattern. -->
    <xsl:template match="wsdl:service" mode="control_pattern_addition">
        <xsl:variable name="service_name" select="@name"/>
        <!-- for each service a control pattern entry for inclusion into the blueprint is created -->
        <pattern_instance image="120"
                          layer_ref="{$Chosen_Project_Layer}"
                          pattern_instance_ref="{$Chosen_Name_Prefix}{translate($service_name, '- ', '_')}"
                          pool_ref="{$Chosen_Pattern_Instance_Pool}"
                          runtime_filename="">

            <!-- Add a splice back to itself to store the fact that this node is a wsdl acess service node
                 Along with the names of the WSDL files involved in creating it -->
            <pattern_splice name="access_service_configuration">
                <source node_ref="Access_Service"/>
                <sink pattern_instance_ref="{$Chosen_Name_Prefix}{translate($service_name, '- ', '_')}"
                      node_ref="Access_Service"
                      action_ref="Invoke"
                      runtime_type="wsdl_access_service"
                      runtime_driver="{/xg:message/xg:header/xg:attributes/xg:attribute[@xg:name='Chosen_WSDL']}"
                      runtime_instance="{$service_name}"/>
            </pattern_splice>

            <!-- Cycle through each of the ports which relates to each seperate node
                 For now we ignore ports that only provide soap 1.2 addresses as this is not supported. -->
            <xsl:for-each select="wsdl:port[*[namespace-uri() != 'http://schemas.xmlsoap.org/wsdl/soap12/']]">
                <!-- this variable is used to match the correct binding with the port being looked at by doing some simple string processing -->
                <xsl:variable name="full_binding_name">
                    <xsl:value-of select="@binding"/>
                </xsl:variable>
                <xsl:variable name="binding_name">
                    <xsl:choose>
                        <xsl:when test="contains(@binding, ':')">
                            <xsl:value-of select="substring-after(@binding,':')"/>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:value-of select="@binding"/>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:variable>
                <!-- parameters are passed with the template call to keep the data that we need -->
                <xsl:apply-templates mode="process_binding" select="/xg:message/xg:body/hy:XStore_Response/xstore:xstore/xstore:content/wsdl:definitions/wsdl:binding[@name = $binding_name]">
                    <xsl:with-param name="servicePort" select="."/>
                    <xsl:with-param name="counter" select="position()"/>
                </xsl:apply-templates>
            </xsl:for-each>
        </pattern_instance>
    </xsl:template>
    <xsl:template match="wsdl:binding" mode="process_binding">
        <xsl:param name="servicePort"/>
        <xsl:param name="counter"/>

        <!-- this cycles through each of the operations which correspond to actions -->
        <xsl:for-each select="wsdl:operation">

            <!-- the action ref is either the soap action if it is available. if it isn't then the operation name is selected -->
            <xsl:variable name="actionName">
                <xsl:choose>
                    <xsl:when test="soap:operation/@soapAction!=''">
                        <xsl:value-of select="soap:operation/@soapAction"/>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:value-of select="@name"/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:variable>

            <!-- check this is not going to give a duplicate action name to one already processed -->
            <xsl:if test="not(preceding-sibling::wsdl:operation[((soap:operation/@soapAction != '') and (soap:operation/@soapAction = $actionName)) or (not(soap:operation/@soapAction != '') and (@name = $actionName))])" >

                <!-- the name of the splice is generated from the name of the operation that we are currently on -->
                <pattern_splice name="{@name}_Splice_{$counter}" desc="Calls the {@name} operation on the remote service">
                    <source node_ref="Access_Service"/>
                    <sink action_ref="{$actionName}">

                        <!-- the node_ref is constructed from the name of the binding from which the operation comes from -->
                        <xsl:attribute name="node_ref">
                            <xsl:value-of select="$Chosen_Name_Prefix"/>
                            <xsl:value-of select="translate($servicePort/@name, '- ', '_')"/>
                        </xsl:attribute>
                        <!-- the pattern_instance_ref points to the proxy and is constructed again from the service name -->
                        <xsl:attribute name="pattern_instance_ref">
                            <xsl:value-of select="$Chosen_Name_Prefix"/>
                            <xsl:value-of select="translate($servicePort/../@name, '- ', '_')"/>
                            <xsl:text>_Proxy</xsl:text>
                        </xsl:attribute>
                        <xsl:attribute name="runtime_instance">
                            <xsl:choose>
                                <xsl:when test="$servicePort/soap:address/@location!=''">
                                    <xsl:value-of select="$servicePort/soap:address/@location"/>
                                </xsl:when>
                                <xsl:when test="$servicePort/http:address/@location!=''">
                                    <xsl:value-of select="$servicePort/http:address/@location"/>
                                </xsl:when>
                                <xsl:when test="$servicePort/soap12:address/@location!=''">
                                    <xsl:value-of select="$servicePort/soap12:address/@location"/>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:text>ERROR no address of web service found</xsl:text>
                                </xsl:otherwise>
                            </xsl:choose>
                        </xsl:attribute>
                        <xsl:attribute name="runtime_protocol">
                            <xsl:text>http</xsl:text>
                        </xsl:attribute>
                        <xsl:attribute name="runtime_type">
                            <xsl:choose>
                                <xsl:when test="../soap:binding">
                                    <xsl:text>soap_service</xsl:text>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:text>http_service</xsl:text>
                                </xsl:otherwise>
                            </xsl:choose>
                        </xsl:attribute>
                    </sink>
                </pattern_splice>
            </xsl:if>
        </xsl:for-each>
    </xsl:template>

</xsl:stylesheet>
