<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:fm="http://www.hyfinity.com/formmaker" xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/" xmlns:exsl="http://exslt.org/common" exclude-result-prefixes="exsl" extension-element-prefixes="exsl">


    <xsl:variable name="opNameSeperator">*@*@*</xsl:variable>

    <xsl:variable name="wsdlFiles" select="/fm:wsdls"/>

    <xsl:template match="/">

        <xsl:variable name="portName" select="substring-before(/fm:wsdls/fm:data_source/@action, $opNameSeperator)"/>
        <xsl:variable name="opName" select="substring-before(substring-after(/fm:wsdls/fm:data_source/@action, $opNameSeperator), $opNameSeperator)"/>
        <xsl:variable name="opNum" select="substring-after(substring-after(/fm:wsdls/fm:data_source/@action, $opNameSeperator), $opNameSeperator)"/>

        <xsl:variable name="matches">
            <fm:matches>
                <xsl:call-template name="find-details" >
                    <xsl:with-param name="wsdl_file" select="$wsdlFiles/fm:file[1]/@name"/>
                    <xsl:with-param name="operation" select="$opName"/>
                    <xsl:with-param name="op_num" select="$opNum"/>
                    <xsl:with-param name="port_type" select="$portName"/>
                    <xsl:with-param name="port_type_ns" select="$wsdlFiles/fm:file[1]/wsdl:definitions/@targetNamespace"/>
                </xsl:call-template>
            </fm:matches>
        </xsl:variable>

        <xsl:variable name="matchesSet" select="exsl:node-set($matches)"/>

        <xsl:choose>
            <xsl:when test="$matchesSet/fm:matches/fm:wsdl_info" >
                <!-- For now we just take the first matching WSDL details.
                     We may need to improve this in the future, eg ignore (or prefer?) SOAP 1.2 details etc. -->
                <xsl:copy-of select="$matchesSet/fm:matches/fm:wsdl_info[1]" />
            </xsl:when>
            <xsl:otherwise>
                <fm:error>Could not determine wsdl info</fm:error>
            </xsl:otherwise>
        </xsl:choose>


    </xsl:template>


    <xsl:template name="find-details">
        <xsl:param name="wsdl_file" />
        <xsl:param name="operation" />
        <xsl:param name="op_num" />
        <xsl:param name="port_type" />
        <xsl:param name="port_type_ns" />
        <xsl:param name="binding" />
        <xsl:param name="binding_ns" />
        <xsl:param name="service" />
        <xsl:param name="service_ns" />


        <xsl:choose>
            <xsl:when test="$binding = ''">
                <!-- Need to find binding detail -->
                <!-- first check if a match can be found in the current file -->
                <xsl:variable name="sameFileBinding">
                    <fm:bindings>
                        <xsl:for-each select="$wsdlFiles/fm:file[@name = $wsdl_file]/wsdl:definitions/wsdl:binding">
                            <xsl:if test="@type = $port_type">
                                <fm:name>
                                    <xsl:value-of select="@name" />
                                </fm:name>
                            </xsl:if>
                            <!-- check namespace prefixes -->
                            <xsl:if test="contains(@type, ':')" >
                                <xsl:variable name="prefix" select="substring-before(@type, ':')"/>
                                <xsl:variable name="value" select="substring-after(@type, ':')"/>
                                <xsl:variable name="uri" select="namespace::*[name() = $prefix]"/>

                                <xsl:if test="($uri = $port_type_ns) and ($value = $port_type)" >
                                    <fm:name>
                                        <xsl:value-of select="@name" />
                                    </fm:name>
                                </xsl:if>
                            </xsl:if>
                        </xsl:for-each>
                    </fm:bindings>
                </xsl:variable>

                <xsl:variable name="sameFileBindingSet" select="exsl:node-set($sameFileBinding)"/>

                <xsl:choose>
                    <xsl:when test="$sameFileBindingSet/fm:bindings/fm:name" >
                        <!-- found at least one binding match, so need to check all of these to see if we find a matching service -->
                        <xsl:for-each select="$sameFileBindingSet/fm:bindings/fm:name">
                            <xsl:call-template name="find-details" >
                                <xsl:with-param name="wsdl_file" select="$wsdl_file"/>
                                <xsl:with-param name="operation" select="$operation"/>
                                <xsl:with-param name="op_num" select="$op_num"/>
                                <xsl:with-param name="port_type" select="$port_type"/>
                                <xsl:with-param name="port_type_ns" select="$port_type_ns"/>
                                <xsl:with-param name="binding" select="."/>
                                <xsl:with-param name="binding_ns" select="$wsdlFiles/fm:file[@name = $wsdl_file]/wsdl:definitions/@targetNamespace"/>
                            </xsl:call-template>
                        </xsl:for-each>
                    </xsl:when>
                    <xsl:otherwise>
                        <!-- No matching bindings found in this file, so look through files that import this one -->
                        <xsl:for-each select="$wsdlFiles/fm:file[wsdl:definitions/wsdl:import/@location = $wsdl_file]" >
                            <xsl:call-template name="find-details" >
                                <xsl:with-param name="wsdl_file" select="@name"/>
                                <xsl:with-param name="operation" select="$operation"/>
                                <xsl:with-param name="op_num" select="$op_num"/>
                                <xsl:with-param name="port_type" select="$port_type"/>
                                <xsl:with-param name="port_type_ns" select="$port_type_ns"/>
                            </xsl:call-template>
                        </xsl:for-each>
                    </xsl:otherwise>
                </xsl:choose>

            </xsl:when>
            <xsl:when test="$service = ''">
                <!-- Need to find service detail -->

                <!-- first check if a match can be found in the current file -->
                <xsl:variable name="sameFileService">
                    <fm:services>
                        <xsl:for-each select="$wsdlFiles/fm:file[@name = $wsdl_file]/wsdl:definitions/wsdl:service/wsdl:port">
                            <xsl:if test="@binding = $binding">
                                <fm:name>
                                    <xsl:value-of select="../@name" />
                                </fm:name>
                            </xsl:if>
                            <!-- check namespace prefixes -->
                            <xsl:if test="contains(@binding, ':')" >
                                <xsl:variable name="prefix" select="substring-before(@binding, ':')"/>
                                <xsl:variable name="value" select="substring-after(@binding, ':')"/>
                                <xsl:variable name="uri" select="namespace::*[name() = $prefix]"/>

                                <xsl:if test="($uri = $binding_ns) and ($value = $binding)" >
                                    <fm:name>
                                        <xsl:value-of select="../@name" />
                                    </fm:name>
                                </xsl:if>
                            </xsl:if>
                        </xsl:for-each>
                    </fm:services>
                </xsl:variable>

                <xsl:variable name="sameFileServiceSet" select="exsl:node-set($sameFileService)"/>

                <xsl:choose>
                    <xsl:when test="$sameFileServiceSet/fm:services/fm:name" >
                        <xsl:for-each select="$sameFileServiceSet/fm:services/fm:name">
                            <xsl:call-template name="find-details" >
                                <xsl:with-param name="wsdl_file" select="$wsdl_file"/>
                                <xsl:with-param name="operation" select="$operation"/>
                                <xsl:with-param name="op_num" select="$op_num"/>
                                <xsl:with-param name="port_type" select="$port_type"/>
                                <xsl:with-param name="port_type_ns" select="$port_type_ns"/>
                                <xsl:with-param name="binding" select="$binding"/>
                                <xsl:with-param name="binding_ns" select="$binding_ns"/>
                                <xsl:with-param name="service" select="."/>
                                <xsl:with-param name="service_ns" select="$wsdlFiles/fm:file[@name = $wsdl_file]/wsdl:definitions/@targetNamespace"/>
                            </xsl:call-template>
                        </xsl:for-each>
                    </xsl:when>
                    <xsl:otherwise>
                        <!-- look through files that import this one -->
                        <xsl:for-each select="$wsdlFiles/fm:file[wsdl:definitions/wsdl:import/@location = $wsdl_file]" >
                            <xsl:call-template name="find-details" >
                                <xsl:with-param name="wsdl_file" select="@name"/>
                                <xsl:with-param name="operation" select="$operation"/>
                                <xsl:with-param name="op_num" select="$op_num"/>
                                <xsl:with-param name="port_type" select="$port_type"/>
                                <xsl:with-param name="port_type_ns" select="$port_type_ns"/>
                                <xsl:with-param name="binding" select="$binding"/>
                                <xsl:with-param name="binding_ns" select="$binding_ns"/>
                            </xsl:call-template>
                        </xsl:for-each>
                    </xsl:otherwise>
                </xsl:choose>


            </xsl:when>
            <xsl:otherwise>
                <!-- all details found so need to output -->
                <fm:wsdl_info root_wsdl="{$wsdl_file}"
                              operation="{$operation}"
                              op_num="{$op_num}"
                              port_type="{$port_type}"
                              port_type_ns="{$port_type_ns}"
                              binding="{$binding}"
                              binding_ns="{$binding_ns}"
                              service="{$service}"
                              service_ns="{$service_ns}"/>
            </xsl:otherwise>
        </xsl:choose>


    </xsl:template>

</xsl:stylesheet>
