<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet version="1.0" xmlns:fm="http://www.hyfinity.com/formmaker" xmlns:xg="http://www.hyfinity.com/xgate" xmlns:xfact="http://www.hyfinity.com/xfactory" xmlns:hy="http://www.hyfinity.com/xengine" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:exsl="http://exslt.org/common" extension-element-prefixes="exsl">

    <!-- find the handler/controller being generated -->
    <xsl:variable name="handler" select="/xg:wrapper/fm:application/fm:handlers/fm:handler[@id = /xg:wrapper/xg:header/xg:gen_control]"/>

    <!-- find the relevant links and pages to check for data sources for this controller
         This should also correctly handle subsection controllers.  For these the links will be to the page not the controller. -->
    <xsl:variable name="incomingActions" select="/xg:wrapper/fm:application/fm:links/fm:link[(fm:target[@type = 'Handler'] = $handler/@id) or ((fm:target[@type = 'Page'] = $handler/@id) and ($handler/@type = 'subsection'))]"/>
    <xsl:variable name="targetPages" select="/xg:wrapper/fm:application/fm:pages/fm:page[@id = /xg:wrapper/fm:application/fm:links/fm:link[(fm:source[@type = 'Handler'] = $handler/@id) and (fm:target/@type = 'Page')]/fm:target] | /xg:wrapper/fm:application/fm:pages/fm:page[(@id = $handler/@id) and ($handler/@type = 'subsection')]"/>

    <xsl:variable name="opNameSeperator">*@*@*</xsl:variable>

    <xsl:variable name="eformDataXPath" >
        <xsl:text>/</xsl:text>
        <xsl:value-of select="$MVC-PREFIX" />
        <xsl:text>:eForm/</xsl:text>
        <xsl:value-of select="$MVC-PREFIX" />
        <xsl:text>:Data</xsl:text>
    </xsl:variable>

    <xsl:variable name="rulebase" select="/xg:wrapper/hy:rulebase"/>

    <xsl:variable name="rulebaseNamespaces">
        <xsl:call-template name="find-namespaces"/>
    </xsl:variable>

    <xsl:variable name="rulebaseNamespacesSet" select="exsl:node-set($rulebaseNamespaces)"/>

    <xsl:variable name="MVC-NAMESPACE">http://www.hyfinity.com/mvc</xsl:variable>
    <xsl:variable name="MVC-PREFIX" select="$rulebaseNamespacesSet/namespaces/namespace[@value =  $MVC-NAMESPACE]/@prefix" />
    <xsl:variable name="DEFAULT-FIELDS-CONTAINER-NAME">formData</xsl:variable>




    <xsl:template match="/">

        <xg:wrapper>
            <xsl:copy-of select="/xg:wrapper/fm:formmaker" />
            <xsl:copy-of select="/xg:wrapper/xfact:project_details" />
            <xsl:copy-of select="/xg:wrapper/xg:header" />

            <!-- Find the list of used services that are not yet used by existing rules.
                 It is possible that these services will be used by the rules we are generating
                 below, so the main point of this processing is to convert the nodes into the
                 four part ids for easy checking in rules against the finished logicsheet. -->
            <xfact:services_to_check>
                <xsl:apply-templates select="/xg:wrapper/xfact:patterns" mode="process_services"/>
            </xfact:services_to_check>

            <!-- Update app map to indicate that the datasources have been processed. -->
            <xsl:apply-templates select="/xg:wrapper/fm:application" mode="process_appmap"/>
            <!-- add the needed rules to the logicsheet -->
            <xsl:apply-templates select="/xg:wrapper/hy:rulebase" mode="process_rulebase"/>

            <!-- Find the list of nodes that need to be spliced to (eg databases) -->
            <xfact:splices_needed>
                <!-- first check new action data sources -->
                <xsl:for-each select="$incomingActions/fm:data_sources/fm:data_source[not(@generated = 'true') and (@type = 'db')]" >
                    <!-- Check if this is already included in the list of services -->
                    <xsl:variable name="patternID" select="substring-before(substring-after(substring-after(@name, '-'), '-'), '-')"/>
                    <xsl:variable name="nodeID" select="substring-after(substring-after(substring-after(@name, '-'), '-'), '-')"/>
                    <xsl:if test="not(/xg:wrapper/xfact:patterns/xfact:pattern[@pattern_id = $patternID]/xfact:nodes/xfact:node[@node_id = $nodeID])" >
                        <xfact:splice pattern="{$patternID}" node="{$nodeID}" action="Query"/>
                    </xsl:if>
                </xsl:for-each>
                <!-- now check new target page data sources -->
                <xsl:for-each select="$targetPages/fm:data_sources/fm:data_source[not(contains(concat(',', @generated, ','), concat(',', $handler/@id, ','))) and (@type = 'db')]" >
                    <!-- Check if this is already included in the list of services -->
                    <xsl:variable name="patternID" select="substring-before(substring-after(substring-after(@name, '-'), '-'), '-')"/>
                    <xsl:variable name="nodeID" select="substring-after(substring-after(substring-after(@name, '-'), '-'), '-')"/>
                    <xsl:if test="not(/xg:wrapper/xfact:patterns/xfact:pattern[@pattern_id = $patternID]/xfact:nodes/xfact:node[@node_id = $nodeID])" >
                        <xfact:splice pattern="{$patternID}" node="{$nodeID}" action="Query"/>
                    </xsl:if>
                </xsl:for-each>

                <!-- Check for wsdl request services -->
                <xsl:for-each select="$incomingActions/fm:data_sources/fm:data_source[not(@generated = 'true') and (@type = 'wsdl') and (@use = 'request')]" >

                    <xsl:variable name="possibleTargets" select="/xg:wrapper/fm:access_service_nodes/xfact:patterns/xfact:pattern/xfact:nodes/xfact:node[(@runtime_driver = current()/fm:wsdl_info/@root_wsdl) and (@runtime_instance = current()/fm:wsdl_info/@service)]"/>
                    <xsl:variable name="existingTargets">
                        <xsl:for-each select="$possibleTargets" >
                            <xsl:if test="/xg:wrapper/xfact:patterns/xfact:pattern[@pattern_id = current()/../../@pattern_id]/xfact:nodes/xfact:node[@node_id = current()/@node_id]" >
                                <xsl:text>match</xsl:text>
                            </xsl:if>
                        </xsl:for-each>
                    </xsl:variable>
                    <xsl:if test="$existingTargets = ''">
                        <xfact:splice pattern="{$possibleTargets[1]/../../@pattern_id}" node="{$possibleTargets[1]/@node_id}" action="{$possibleTargets[1]/xfact:actions/xfact:action[1]/@alias}"/>
                    </xsl:if>

                </xsl:for-each>


            </xfact:splices_needed>

        </xg:wrapper>

    </xsl:template>

    <!-- deterine whether there are any existing rules calling this node,
         and if not ouput the id to check again in the rules later. -->
    <xsl:template match="xfact:node" mode="process_services">
        <xsl:variable name="fullID" >
            <xsl:value-of select="/xg:wrapper/xfact:project_details/xfact:product" />
            <xsl:text>-</xsl:text>
            <xsl:value-of select="/xg:wrapper/xfact:project_details/xfact:project" />
            <xsl:text>-</xsl:text>
            <xsl:value-of select="../../@pattern_id" />
            <xsl:text>-</xsl:text>
            <xsl:value-of select="@node_id" />
        </xsl:variable>

        <xsl:if test="not(/xg:wrapper/hy:rulebase/hy:rule/hy:head/hy:result/hy:target[@name = $fullID])" >
            <xfact:service action="{xfact:actions/xfact:action[1]/@alias}">
                <xsl:value-of select="$fullID" />
            </xfact:service>
        </xsl:if>
    </xsl:template>



    <!-- Mark the data sources that we are now generating to stop us doing it again in the future. -->
    <xsl:template match="fm:data_source" mode="process_appmap">
        <xsl:copy>
            <xsl:copy-of select="@*"/>

            <!-- check if this data source is one of the incoming action data sources that we will process -->
            <xsl:if test="count($incomingActions/fm:data_sources/fm:data_source | current()) = count($incomingActions/fm:data_sources/fm:data_source)" >
                <xsl:attribute name="generated">true</xsl:attribute>
            </xsl:if>

            <!-- check if this data source is one of the target page data sources that we will process -->
            <xsl:if test="count($targetPages/fm:data_sources/fm:data_source | current()) = count($targetPages/fm:data_sources/fm:data_source)" >
                <!-- There could be multiple controllers that link to this page, so rather than using a true
                     or false value for the generated flag, we instead store the ids of all the handlers that
                     have been generated so far. -->
                <xsl:if test="not(contains(concat(',', @generated, ','), concat(',', $handler/@id, ',')))">
                    <xsl:attribute name="generated">
                        <xsl:value-of select="@generated"/>
                        <xsl:if test="@generated != ''">,</xsl:if>
                        <xsl:value-of select="$handler/@id" />
                    </xsl:attribute>
                </xsl:if>
            </xsl:if>

        </xsl:copy>
    </xsl:template>

    <xsl:template match="node() | @*" mode="process_appmap">
        <xsl:copy>
            <xsl:apply-templates select="@*" mode="process_appmap"/>
            <xsl:apply-templates mode="process_appmap"/>
        </xsl:copy>
    </xsl:template>


    <xsl:template match="hy:rulebase" mode="process_rulebase">
        <temp>
            <!--<xsl:copy>
            <xsl:copy-of select="@*"/>-->

            <!-- Handle namespaces  It is possible that the xpaths generated for each data source
                 use namespace prefixes that are not defined in this logicsheet!!!!! -->

            <xsl:text disable-output-escaping="yes">&lt;hy:rulebase xmlns:hy="http://www.hyfinity.com/xengine"</xsl:text>
            <!-- output all the existing attributes -->
            <xsl:for-each select="@*">
                <xsl:text> </xsl:text>
                <xsl:value-of select="name()"/>
                <xsl:text disable-output-escaping="yes">="</xsl:text>
                <xsl:value-of select="."/>
                <xsl:text disable-output-escaping="yes">"</xsl:text>
            </xsl:for-each>

            <!-- output all the namespaces (both new and existing)-->
            <xsl:for-each select="$rulebaseNamespacesSet/namespaces/namespace" >
                <xsl:if test="@prefix != 'hy'">
                    <xsl:text> xmlns:</xsl:text>
                    <xsl:value-of select="@prefix"/>
                    <xsl:text disable-output-escaping="yes">="</xsl:text>
                    <xsl:value-of select="@value"/>
                    <xsl:text disable-output-escaping="yes">"</xsl:text>
                </xsl:if>
            </xsl:for-each>

            <xsl:text disable-output-escaping="yes">></xsl:text>


            <!-- copy all existing rules etc -->
            <xsl:copy-of select="*"/>

            <!-- generate any new rules -->


            <!-- First check the incoming links - ie wsdl calls or sql inserts/updates -->
            <xsl:for-each select="$incomingActions/fm:data_sources/fm:data_source[not(@generated = 'true')]" >
                <xsl:choose>
                    <xsl:when test="@type = 'wsdl' and @use = 'request'">
                        <xsl:call-template name="generate-wsdl-rule">
                            <xsl:with-param name="datasource" select="."/>
                        </xsl:call-template>
                    </xsl:when>
                    <xsl:when test="@type = 'db' and contains(@use, 'edit')">
                        <xsl:choose>
                            <xsl:when test="@sqlOperation = 'update'">
                                <xsl:call-template name="generate-sql-update-rule">
                                    <xsl:with-param name="datasource" select="."/>
                                </xsl:call-template>
                            </xsl:when>
                            <xsl:when test="@sqlOperation = 'insert'">
                                <xsl:call-template name="generate-sql-insert-rule">
                                    <xsl:with-param name="datasource" select="."/>
                                </xsl:call-template>
                            </xsl:when>
                            <xsl:when test="@sqlOperation = 'delete'">
                                <xsl:call-template name="generate-sql-delete-rule">
                                    <xsl:with-param name="datasource" select="."/>
                                </xsl:call-template>
                            </xsl:when>
                        </xsl:choose>
                    </xsl:when>
                </xsl:choose>

            </xsl:for-each>


            <!-- Check if we need to add a sql commit statement
                 This is done if the set of new data sources include maintenance database calls,
                 and the logicsheet does not already contain any commit statements. -->
            <xsl:for-each select="$incomingActions/fm:data_sources/fm:data_source[not(@generated = 'true') and (@type = 'db') and contains(@use, 'edit') and not(@name = preceding::fm:link/fm:data_sources/fm:data_source[not(@generated = 'true') and (@type = 'db') and contains(@use, 'edit')]/@name)]" >
                <!-- now check no commit already present for this database -->
                <xsl:if test="not(/xg:wrapper/hy:rulebase/hy:rule/hy:head/hy:result[hy:operator = 'SQL']/hy:target[@name = current()/@name]/hy:params/hy:param[(@name = 'sql_statement') and (contains(., 'COMMIT') or contains(., 'commit'))])" >
                    <xsl:call-template name="generate-sql-commit-rule">
                        <xsl:with-param name="datasource" select="."/>
                    </xsl:call-template>
                </xsl:if>
            </xsl:for-each>


            <!-- Now check for any target page data sources that need rules added to this controller (ie sql selects) -->
            <xsl:for-each select="$targetPages/fm:data_sources/fm:data_source[not(contains(concat(',', @generated, ','), concat(',', $handler/@id, ',')))]" >
                <xsl:choose>
                    <xsl:when test="@type = 'db' and contains(@use, 'Many')">
                        <xsl:call-template name="generate-sql-select-all-rule">
                            <xsl:with-param name="datasource" select="."/>
                        </xsl:call-template>
                    </xsl:when>
                    <xsl:when test="@type = 'db'">
                        <xsl:call-template name="generate-sql-select-one-rule">
                            <xsl:with-param name="datasource" select="."/>
                        </xsl:call-template>
                    </xsl:when>
                </xsl:choose>
            </xsl:for-each>

            <xsl:text disable-output-escaping="yes">&lt;/hy:rulebase></xsl:text>
        <!--</xsl:copy>-->
        </temp>

    </xsl:template>

    <xsl:template name="generate-wsdl-rule" xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/" xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/">
        <xsl:param name="datasource" />

        <xsl:variable name="possibleTargets" select="/xg:wrapper/fm:access_service_nodes/xfact:patterns/xfact:pattern/xfact:nodes/xfact:node[(@runtime_driver = $datasource/fm:wsdl_info/@root_wsdl) and (@runtime_instance = $datasource/fm:wsdl_info/@service)]"/>
        <xsl:variable name="existingTargets">
            <xsl:for-each select="$possibleTargets" >
                <xsl:if test="/xg:wrapper/xfact:patterns/xfact:pattern[@pattern_id = current()/../../@pattern_id]/xfact:nodes/xfact:node[@node_id = current()/@node_id]" >
                    <xsl:value-of select="../../@pattern_id"/>
                    <xsl:text>-</xsl:text>
                    <xsl:value-of select="@node_id"/>
                    <xsl:text>,</xsl:text>
                </xsl:if>
            </xsl:for-each>
        </xsl:variable>
        <xsl:variable name="selectedNode">
            <xsl:choose>
                <xsl:when test="$existingTargets != ''">
                    <xsl:value-of select="substring-after(substring-before($existingTargets, ','), '-')" />
                </xsl:when>
                <xsl:otherwise>
                    <xsl:value-of select="$possibleTargets[1]/@node_id" />
                </xsl:otherwise>
            </xsl:choose>
        </xsl:variable>
        <xsl:variable name="selectedPattern">
            <xsl:choose>
                <xsl:when test="$existingTargets != ''">
                    <xsl:value-of select="substring-before(substring-before($existingTargets, ','), '-')" />
                </xsl:when>
                <xsl:otherwise>
                    <xsl:value-of select="$possibleTargets[1]/../../@pattern_id" />
                </xsl:otherwise>
            </xsl:choose>
        </xsl:variable>

        <hy:rule id="Call_{$datasource/fm:wsdl_info/@operation}_operation_for_action_{$datasource/../../fm:action}" desc="" disable="false" priority="20" base="pattern">
            <xsl:if test="$selectedNode = ''" >
                <xsl:attribute name="disable">true</xsl:attribute>
                <xsl:attribute name="desc">Please specify the Service to invoke before enabling this rule</xsl:attribute>
            </xsl:if>
            <hy:head>
                <hy:result id="call_service">
                    <hy:operator>Service</hy:operator>
                    <hy:target action="" name="" update="append">
                        <xsl:if test="$selectedNode != ''" >
                            <xsl:attribute name="name">
                                <xsl:value-of select="/xg:wrapper/xfact:project_details/xfact:product" />
                                <xsl:text>-</xsl:text>
                                <xsl:value-of select="/xg:wrapper/xfact:project_details/xfact:project" />
                                <xsl:text>-</xsl:text>
                                <xsl:value-of select="$selectedPattern" />
                                <xsl:text>-</xsl:text>
                                <xsl:value-of select="$selectedNode" />
                            </xsl:attribute>
                            <xsl:attribute name="action">
                                <xsl:value-of select="/xg:wrapper/fm:access_service_nodes/xfact:patterns/xfact:pattern[@pattern_id = $selectedPattern]/xfact:nodes/xfact:node[@node_id = $selectedNode]/xfact:actions/xfact:action[1]/@alias" />
                            </xsl:attribute>
                        </xsl:if>
                        <hy:params>
                            <hy:param name="from_location" type="com.hyfinity.xpath.XPath">
                                <xsl:choose>
                                    <xsl:when test="$datasource/@location[. != '']" >
                                        <xsl:call-template name="output-location-xpath">
                                            <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                                        </xsl:call-template>
                                    </xsl:when>
                                    <xsl:otherwise>
                                        <xsl:value-of select="$eformDataXPath" />
                                        <xsl:text>/*</xsl:text>
                                    </xsl:otherwise>
                                </xsl:choose>
                            </hy:param>
                            <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                            </hy:param>
                        </hy:params>
                    </hy:target>
                </hy:result>
            </hy:head>
            <hy:body>
                <xsl:call-template name="generate-rule-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </hy:body>
        </hy:rule>

    </xsl:template>


    <xsl:template name="generate-sql-update-rule">
        <xsl:param name="datasource" />

        <hy:rule id="Update_{$datasource/@action}_database_table_for_action_{$datasource/../../fm:action}" desc="" disable="false" priority="20" base="pattern">
            <hy:head>
                <hy:result id="update_record">
                <hy:operator>SQL</hy:operator>
                <hy:target action="Query" name="{$datasource/@name}" update="append">
                    <xsl:if test="$datasource/@location[. != '']" >
                        <xsl:attribute name="update">replace</xsl:attribute>
                    </xsl:if>
                    <hy:params>
                        <hy:param name="sql_statement" type="java.lang.String">
                            <xsl:text>UPDATE </xsl:text>
                            <xsl:value-of select="$datasource/@action" />
                            <xsl:text> SET </xsl:text>
                            <xsl:call-template name="create-non-primary-key-set">
                                <xsl:with-param name="column-list" select="$datasource/columns"/>
                            </xsl:call-template>
                            <xsl:text>
    WHERE </xsl:text>
                            <xsl:call-template name="create-primary-key-where-clause" >
                                <xsl:with-param name="column-list" select="$datasource/columns"/>
                            </xsl:call-template>
                            <xsl:text>;</xsl:text>
                        </hy:param>

                        <!-- QUESTION: Do we need to format the response at all?? -->
                        <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                            <xsl:choose>
                                <xsl:when test="$datasource/@location[. != '']">
                                    <xsl:choose>
                                        <!-- In an editMany scenario this action will be in a looping rule,
                                             so want to replace just the value updated, not the whole request. -->
                                        <xsl:when test="$datasource/@use = 'editMany'" >
                                            <xsl:text>.</xsl:text>
                                        </xsl:when>
                                        <xsl:otherwise>
                                            <xsl:call-template name="output-location-xpath">
                                                <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                                            </xsl:call-template>
                                        </xsl:otherwise>
                                    </xsl:choose>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:value-of select="$eformDataXPath" />
                                </xsl:otherwise>
                            </xsl:choose>
                        </hy:param>

                        <xsl:variable name="contextXPath">
                            <xsl:if test="$datasource/@use != 'editMany'" >
                                <xsl:choose>
                                    <xsl:when test="$datasource/@location">
                                        <xsl:call-template name="output-location-xpath">
                                            <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                                        </xsl:call-template>
                                    </xsl:when>
                                    <xsl:otherwise>
                                        <xsl:value-of select="$eformDataXPath" />
                                    </xsl:otherwise>
                                </xsl:choose>
                                <xsl:text>/record</xsl:text>
                            </xsl:if>
                        </xsl:variable>

                        <xsl:call-template name="create-primary-key-params">
                            <xsl:with-param name="column-list" select="$datasource/columns"/>
                            <xsl:with-param name="container-xpath" select="$contextXPath"/>
                        </xsl:call-template>

                        <xsl:call-template name="create-non-primary-key-params">
                            <xsl:with-param name="column-list" select="$datasource/columns"/>
                            <xsl:with-param name="container-xpath" select="$contextXPath"/>
                        </xsl:call-template>

                    </hy:params>
                </hy:target>
            </hy:result>

            </hy:head>
            <hy:body>
                <xsl:call-template name="generate-rule-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </hy:body>
        </hy:rule>

    </xsl:template>

    <xsl:template name="generate-sql-insert-rule">
        <xsl:param name="datasource" />

        <hy:rule id="Insert_{$datasource/@action}_database_entry_for_action_{$datasource/../../fm:action}" desc="" disable="false" priority="20" base="pattern">
            <hy:head>
                <hy:result id="insert_record">
                <hy:operator>SQL</hy:operator>
                <hy:target action="Query" name="{$datasource/@name}" update="append">
                    <xsl:if test="$datasource/@location[. != '']" >
                        <xsl:attribute name="update">replace</xsl:attribute>
                    </xsl:if>
                    <hy:params>
                        <hy:param name="sql_statement" type="java.lang.String">
                            <xsl:text>INSERT INTO </xsl:text>
                            <xsl:value-of select="$datasource/@action" />
                            <xsl:text>(</xsl:text>
                            <xsl:call-template name="create-field-list">
                                <xsl:with-param name="column-list" select="$datasource/columns"/>
                                <xsl:with-param name="ignoreAutoGen">true</xsl:with-param>
                            </xsl:call-template>
                            <xsl:text>)
    VALUES (</xsl:text>
                            <xsl:call-template name="create-field-list">
                                <xsl:with-param name="column-list" select="$datasource/columns"/>
                                <xsl:with-param name="ignoreAutoGen">true</xsl:with-param>
                                <xsl:with-param name="prefix">$</xsl:with-param>
                            </xsl:call-template>
                            <xsl:text>);</xsl:text>
                        </hy:param>


                        <!-- QUESTION: Do we need to format the response at all?? -->
                        <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                            <xsl:choose>
                                <xsl:when test="$datasource/@location[. != '']">
                                    <xsl:choose>
                                        <!-- In an editMany scenario this action will be in a looping rule,
                                             so want to replace just the value updated, not the whole request. -->
                                        <xsl:when test="$datasource/@use = 'editMany'" >
                                            <xsl:text>.</xsl:text>
                                        </xsl:when>
                                        <xsl:otherwise>
                                            <xsl:call-template name="output-location-xpath">
                                                <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                                            </xsl:call-template>
                                        </xsl:otherwise>
                                    </xsl:choose>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:value-of select="$eformDataXPath" />
                                </xsl:otherwise>
                            </xsl:choose>
                        </hy:param>

                        <xsl:variable name="contextXPath">
                            <xsl:if test="$datasource/@use != 'editMany'" >
                                <xsl:choose>
                                    <xsl:when test="$datasource/@location">
                                        <xsl:call-template name="output-location-xpath">
                                            <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                                        </xsl:call-template>
                                    </xsl:when>
                                    <xsl:otherwise>
                                        <xsl:value-of select="$eformDataXPath" />
                                    </xsl:otherwise>
                                </xsl:choose>
                                <xsl:text>/record</xsl:text>
                            </xsl:if>
                        </xsl:variable>

                        <xsl:call-template name="create-primary-key-params">
                            <xsl:with-param name="column-list" select="$datasource/columns"/>
                            <xsl:with-param name="container-xpath" select="$contextXPath"/>
                            <xsl:with-param name="ignoreAutoGen">true</xsl:with-param>
                        </xsl:call-template>

                        <xsl:call-template name="create-non-primary-key-params">
                            <xsl:with-param name="column-list" select="$datasource/columns"/>
                            <xsl:with-param name="container-xpath" select="$contextXPath"/>
                        </xsl:call-template>

                    </hy:params>
                </hy:target>
            </hy:result>

            </hy:head>
            <hy:body>
                <xsl:call-template name="generate-rule-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </hy:body>
        </hy:rule>

    </xsl:template>

    <xsl:template name="generate-sql-delete-rule">
        <xsl:param name="datasource" />

        <hy:rule id="Delete_{$datasource/@action}_database_entry_for_action_{$datasource/../../fm:action}" desc="" disable="false" priority="20" base="pattern">
            <hy:head>
                <hy:result id="delete_record">
                <hy:operator>SQL</hy:operator>
                <hy:target action="Query" name="{$datasource/@name}" update="append">
                    <xsl:if test="$datasource/@location[. != '']" >
                        <xsl:attribute name="update">replace</xsl:attribute>
                    </xsl:if>
                    <hy:params>
                        <hy:param name="sql_statement" type="java.lang.String">
                            <xsl:text>DELETE FROM </xsl:text>
                            <xsl:value-of select="$datasource/@action" />
                            <xsl:text> WHERE </xsl:text>
                            <xsl:call-template name="create-primary-key-where-clause" >
                                <xsl:with-param name="column-list" select="$datasource/columns"/>
                            </xsl:call-template>
                            <xsl:text>;</xsl:text>
                        </hy:param>

                        <!-- QUESTION: Do we need to format the response at all?? -->
                        <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                            <xsl:choose>
                                <xsl:when test="$datasource/@location[. != '']">
                                    <xsl:choose>
                                        <!-- In an editMany scenario this action will be in a looping rule,
                                             so want to replace just the value updated, not the whole request. -->
                                        <xsl:when test="$datasource/@use = 'editMany'" >
                                            <xsl:text>.</xsl:text>
                                        </xsl:when>
                                        <xsl:otherwise>
                                            <xsl:call-template name="output-location-xpath">
                                                <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                                            </xsl:call-template>
                                        </xsl:otherwise>
                                    </xsl:choose>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:value-of select="$eformDataXPath" />
                                </xsl:otherwise>
                            </xsl:choose>
                        </hy:param>

                        <xsl:variable name="contextXPath">
                            <xsl:if test="$datasource/@use != 'editMany'" >
                                <xsl:choose>
                                    <xsl:when test="$datasource/@location">
                                        <xsl:call-template name="output-location-xpath">
                                            <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                                        </xsl:call-template>
                                    </xsl:when>
                                    <xsl:otherwise>
                                        <xsl:value-of select="$eformDataXPath" />
                                    </xsl:otherwise>
                                </xsl:choose>
                                <xsl:text>/record</xsl:text>
                            </xsl:if>
                        </xsl:variable>

                        <xsl:call-template name="create-primary-key-params">
                            <xsl:with-param name="column-list" select="$datasource/columns"/>
                            <xsl:with-param name="container-xpath" select="$contextXPath"/>
                        </xsl:call-template>
                    </hy:params>
                </hy:target>
            </hy:result>

            </hy:head>
            <hy:body>
                <xsl:call-template name="generate-rule-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </hy:body>
        </hy:rule>

    </xsl:template>

    <xsl:template name="generate-sql-select-all-rule">
        <xsl:param name="datasource" />

        <hy:rule id="Retrieve_all_{$datasource/@action}_database_entries" desc="" disable="false" priority="20" base="pattern">
            <hy:head>
                <hy:result id="retrieve_records">
                    <hy:operator>SQL</hy:operator>
                    <hy:target action="Query" name="{$datasource/@name}" update="append">
                        <hy:params>
                            <hy:param name="sql_statement" type="java.lang.String">
                                <xsl:text>SELECT * FROM </xsl:text>
                                <xsl:value-of select="$datasource/@action" />
                                <xsl:text>;</xsl:text>
                            </hy:param>
                            <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                            </hy:param>
                        </hy:params>
                    </hy:target>
                </hy:result>
                <!-- Format the response to match that needed by the screen!! -->
                <hy:result id="rename_container">
                    <hy:operator>Rename</hy:operator>
                    <hy:target action="rename" name="factbase" update="">
                        <hy:params>
                            <hy:param name="location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                                <xsl:text>/sql_result[last()]/statement</xsl:text>
                            </hy:param>
                            <hy:param name="new_name" type="java.lang.String">
                                <xsl:value-of select="$datasource/@action" />
                            </hy:param>
                            <hy:param name="new_namespace" type="java.lang.String"></hy:param>
                        </hy:params>
                    </hy:target>
                </hy:result>
                <hy:result id="move_container">
                    <hy:operator>Move</hy:operator>
                    <hy:target action="add" name="factbase" update="replace">
                        <hy:params>
                            <hy:param name="from_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                                <xsl:text>/sql_result[last()]/</xsl:text>
                                <xsl:value-of select="$datasource/@action" />
                            </hy:param>
                            <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                                <xsl:text>/sql_result[last()]</xsl:text>
                            </hy:param>
                        </hy:params>
                    </hy:target>
                </hy:result>
            </hy:head>
            <hy:body>
                <xsl:call-template name="generate-rule-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </hy:body>
        </hy:rule>

    </xsl:template>

    <xsl:template name="generate-sql-select-one-rule">
        <xsl:param name="datasource" />

        <hy:rule id="Retrieve_{$datasource/@action}_database_entry" desc="" disable="false" priority="20" base="pattern">
            <hy:head>
                <hy:result id="retrieve_record">
                    <hy:operator>SQL</hy:operator>
                    <hy:target action="Query" name="{$datasource/@name}" update="append">
                        <hy:params>
                            <hy:param name="sql_statement" type="java.lang.String">
                                <xsl:text>SELECT * FROM </xsl:text>
                                <xsl:value-of select="$datasource/@action" />
                                <xsl:text> WHERE </xsl:text>
                                <xsl:call-template name="create-primary-key-where-clause" >
                                    <xsl:with-param name="column-list" select="$datasource/columns"/>
                                </xsl:call-template>
                                <xsl:text>;</xsl:text>
                            </hy:param>
                            <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                            </hy:param>
                            <xsl:call-template name="create-primary-key-params">
                                <xsl:with-param name="column-list" select="$datasource/columns"/>
                                <xsl:with-param name="container-xpath">
                                    <xsl:value-of select="$eformDataXPath" />
                                    <xsl:text>/</xsl:text>
                                    <xsl:value-of select="$MVC-PREFIX" />
                                    <xsl:text>:</xsl:text>
                                    <xsl:value-of select="$DEFAULT-FIELDS-CONTAINER-NAME" />
                                </xsl:with-param>
                                <xsl:with-param name="namespace-prefix" select="$MVC-PREFIX"/>
                            </xsl:call-template>
                        </hy:params>
                    </hy:target>
                </hy:result>
                <!-- Format the response to match that needed by the screen!! -->
                <hy:result id="rename_container">
                    <hy:operator>Rename</hy:operator>
                    <hy:target action="rename" name="factbase" update="">
                        <hy:params>
                            <hy:param name="location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                                <xsl:text>/sql_result[last()]/statement</xsl:text>
                            </hy:param>
                            <hy:param name="new_name" type="java.lang.String">
                                <xsl:value-of select="$datasource/@action" />
                            </hy:param>
                            <hy:param name="new_namespace" type="java.lang.String"></hy:param>
                        </hy:params>
                    </hy:target>
                </hy:result>
                <hy:result id="move_container">
                    <hy:operator>Move</hy:operator>
                    <hy:target action="add" name="factbase" update="replace">
                        <hy:params>
                            <hy:param name="from_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                                <xsl:text>/sql_result[last()]/</xsl:text>
                                <xsl:value-of select="$datasource/@action" />
                            </hy:param>
                            <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                                <xsl:text>/sql_result[last()]</xsl:text>
                            </hy:param>
                        </hy:params>
                    </hy:target>
                </hy:result>
            </hy:head>
            <hy:body>
                <xsl:call-template name="generate-rule-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </hy:body>
        </hy:rule>

    </xsl:template>

    <xsl:template name="generate-sql-commit-rule">
        <xsl:param name="datasource" />

        <xsl:variable name="dbName" select="substring-before(substring-after(substring-after($datasource/@name, '-'), '-'), '-')"/>

        <hy:rule id="commit_database_modifications_to_{$dbName}" desc="Commits all the changes made to this database in the previous rules" disable="false" priority="25" base="pattern">
            <hy:head>
                <hy:result id="delete_record">
                    <hy:operator>SQL</hy:operator>
                    <hy:target action="Query" name="{$datasource/@name}" update="append">
                        <hy:params>
                            <hy:param name="sql_statement" type="java.lang.String">
                                <xsl:text>COMMIT;</xsl:text>
                            </hy:param>
                            <!-- QUESTION: Do we need to format the response at all?? -->
                            <hy:param name="to_location" type="com.hyfinity.xpath.XPath">
                                <xsl:value-of select="$eformDataXPath" />
                            </hy:param>
                        </hy:params>
                    </hy:target>
                </hy:result>
            </hy:head>
            <hy:body>
                <hy:clause id="check_no_errors" multiple="false">
                    <hy:operator>Read</hy:operator>
                    <hy:target action="eval" name="factbase">
                        <hy:params>
                            <hy:param name="xpath_expression" type="com.hyfinity.xpath.XPath">
                                <xsl:text>not(</xsl:text>
                                <xsl:value-of select="$eformDataXPath" />
                                <xsl:text>//sql_result/statement/status[@outcome = 'fail'])</xsl:text>
                            </hy:param>
                        </hy:params>
                    </hy:target>
                </hy:clause>
            </hy:body>
        </hy:rule>

    </xsl:template>


    <!-- outputs a string containing the clause part of a SQL statement
         This is everything after the WHERE bit needed to uniquely identify a record based on primary key.-->
    <xsl:template name="create-primary-key-where-clause">
        <xsl:param name="column-list"/>

        <xsl:for-each select="$column-list/primary_keys/record">
            <xsl:if test="position() &gt; 1">
                <xsl:text> AND </xsl:text>
            </xsl:if>
            <xsl:value-of select="COLUMN_NAME" />
            <xsl:text> = $</xsl:text>
            <xsl:value-of select="COLUMN_NAME" />
        </xsl:for-each>
    </xsl:template>

    <!-- outputs the set of param tags needed within a SQL rule when using the clause output from the above template -->
    <xsl:template name="create-primary-key-params">
        <xsl:param name="column-list"/>
        <xsl:param name="container-xpath" />
        <xsl:param name="namespace-prefix" />
        <xsl:param name="ignoreAutoGen">false</xsl:param>

        <xsl:for-each select="$column-list/list/record[(COLUMN_NAME = $column-list/primary_keys/record/COLUMN_NAME) and (($ignoreAutoGen = 'false') or not(IS_AUTOINCREMENT = 'YES'))]">
            <hy:param name="${COLUMN_NAME}">
                <xsl:call-template name="check-type">
                    <xsl:with-param name="fieldDetails" select="."/>
                </xsl:call-template>
                <xsl:value-of select="$container-xpath" />
                <xsl:if test="$container-xpath != ''">
                    <xsl:text>/</xsl:text>
                </xsl:if>
                <xsl:if test="$namespace-prefix != ''">
                    <xsl:value-of select="$namespace-prefix" />
                    <xsl:text>:</xsl:text>
                </xsl:if>
                <xsl:value-of select="COLUMN_NAME" />
            </hy:param>
        </xsl:for-each>
    </xsl:template>

    <!-- outputs a string containing the SET part of a SQL statement
         This is used to update an existing record -->
    <xsl:template name="create-non-primary-key-set">
        <xsl:param name="column-list"/>

        <xsl:for-each select="$column-list/list/record[not(COLUMN_NAME = $column-list/primary_keys/record/COLUMN_NAME)]">
            <xsl:if test="position() &gt; 1">
                <xsl:text>, </xsl:text>
            </xsl:if>
            <xsl:value-of select="COLUMN_NAME" />
            <xsl:text> = $</xsl:text>
            <xsl:value-of select="COLUMN_NAME" />
        </xsl:for-each>
    </xsl:template>

    <!-- outputs a string containing a comma seperated list of all the fields in the table.
         Optionally adding a prefix to each field name -->
    <xsl:template name="create-field-list">
        <xsl:param name="column-list"/>
        <xsl:param name="prefix" />
        <xsl:param name="ignoreAutoGen">false</xsl:param>

        <xsl:for-each select="$column-list/list/record[($ignoreAutoGen = 'false') or not((COLUMN_NAME = $column-list/primary_keys/record/COLUMN_NAME) and (IS_AUTOINCREMENT = 'YES'))]">

            <xsl:if test="position() &gt; 1">
                <xsl:text>, </xsl:text>
            </xsl:if>

            <xsl:value-of select="$prefix" />
            <xsl:value-of select="COLUMN_NAME" />
        </xsl:for-each>
    </xsl:template>

    <!-- outputs the set of param tags needed within a SQL rule when using the SET fragment output from the above template -->
    <xsl:template name="create-non-primary-key-params">
        <xsl:param name="column-list"/>
        <xsl:param name="container-xpath" />
        <xsl:param name="namespace-prefix" />

        <xsl:for-each select="$column-list/list/record[not(COLUMN_NAME = $column-list/primary_keys/record/COLUMN_NAME)]">
            <hy:param name="${COLUMN_NAME}">

                <xsl:call-template name="check-type">
                    <xsl:with-param name="fieldDetails" select="."/>
                </xsl:call-template>

                <xsl:value-of select="$container-xpath" />
                <xsl:if test="$container-xpath != ''">
                    <xsl:text>/</xsl:text>
                </xsl:if>
                <xsl:if test="$namespace-prefix != ''">
                    <xsl:value-of select="$namespace-prefix" />
                    <xsl:text>:</xsl:text>
                </xsl:if>
                <xsl:value-of select="COLUMN_NAME" />
            </hy:param>
        </xsl:for-each>
    </xsl:template>

    <!-- Outputs a type attribute which converts the database type of the given column to the
         applicable sql param type-->
    <xsl:template name="check-type">
        <xsl:param name="fieldDetails"/>

        <xsl:attribute name="type">
            <!-- These numbers match those defined in the java.sql.Types class.
                 See the following urls for details:
                 http://java.sun.com/javase/6/docs/api/java/sql/DatabaseMetaData.html#getColumns(java.lang.String, java.lang.String, java.lang.String, java.lang.String)
                 http://java.sun.com/javase/6/docs/api/constant-values.html#java.sql.Types.ARRAY -->
            <xsl:choose>
                <xsl:when test="($fieldDetails/DATA_TYPE = '4')">int</xsl:when> <!-- INTEGER -->
                <xsl:when test="($fieldDetails/DATA_TYPE = '6')">float</xsl:when> <!-- FLOAT -->
                <xsl:when test="($fieldDetails/DATA_TYPE = '8')">double</xsl:when> <!-- DOUBLE -->
                <xsl:when test="($fieldDetails/DATA_TYPE = '2')">int</xsl:when> <!-- NUMERIC -->
                <xsl:when test="($fieldDetails/DATA_TYPE = '91')">date</xsl:when> <!-- DATE -->
                <xsl:when test="($fieldDetails/DATA_TYPE = '92')">time</xsl:when> <!-- TIME -->
                <xsl:when test="($fieldDetails/DATA_TYPE = '93')">timestamp</xsl:when> <!-- TIMESTAMP -->
                <xsl:otherwise>string</xsl:otherwise>
            </xsl:choose>
        </xsl:attribute>
    </xsl:template>






    <!-- Generates the condition part of a rule for the given data source -->
    <xsl:template name="generate-rule-condition">
        <xsl:param name="datasource" />

        <xsl:variable name="isLink">
            <xsl:choose>
                <xsl:when test="local-name($datasource/../..) = 'link'" >true</xsl:when>
                <xsl:otherwise>false</xsl:otherwise>
            </xsl:choose>
        </xsl:variable>

        <xsl:variable name="hasLocation">
            <xsl:choose>
                <xsl:when test="$datasource/@location != ''" >true</xsl:when>
                <xsl:otherwise>false</xsl:otherwise>
            </xsl:choose>
        </xsl:variable>

        <xsl:choose>
            <xsl:when test="$isLink = 'true' and $hasLocation = 'true'" >
                <hy:and>
                    <xsl:call-template name="output-action-condition">
                        <xsl:with-param name="datasource" select="$datasource"/>
                    </xsl:call-template>
                    <xsl:call-template name="output-location-condition">
                        <xsl:with-param name="datasource" select="$datasource"/>
                    </xsl:call-template>
                </hy:and>
            </xsl:when>
            <xsl:when test="$isLink = 'true'" >
                <xsl:call-template name="output-action-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </xsl:when>
            <xsl:when test="$hasLocation = 'true'" >
                <xsl:call-template name="output-location-condition">
                    <xsl:with-param name="datasource" select="$datasource"/>
                </xsl:call-template>
            </xsl:when>
        </xsl:choose>
    </xsl:template>


    <xsl:template name="output-action-condition">
        <xsl:param name="datasource" />

        <hy:clause id="check_incoming_action" multiple="false">
            <hy:operator>Read</hy:operator>
            <hy:target action="eval" name="factbase">
                <hy:params>
                    <hy:param name="xpath_expression" type="com.hyfinity.xpath.XPath">
                        <xsl:text>/</xsl:text>
                        <xsl:value-of select="$MVC-PREFIX" />
                        <xsl:text>:eForm/</xsl:text>
                        <xsl:value-of select="$MVC-PREFIX" />
                        <xsl:text>:Control/</xsl:text>
                        <xsl:value-of select="$MVC-PREFIX" />
                        <xsl:text>:action[. = '</xsl:text>
                        <xsl:value-of select="$datasource/../../fm:action" />
                        <xsl:text>']</xsl:text>
                    </hy:param>
                </hy:params>
            </hy:target>
        </hy:clause>

    </xsl:template>

    <xsl:template name="output-location-condition">
        <xsl:param name="datasource" />

        <hy:clause id="check_incoming_data" multiple="false">
            <xsl:if test="$datasource/@type = 'db' and $datasource/@use = 'editMany'" >
                <xsl:attribute name="multiple">true</xsl:attribute>
                <xsl:attribute name="id">for_each_incoming_record</xsl:attribute>
            </xsl:if>
            <hy:operator>Read</hy:operator>
            <hy:target action="eval" name="factbase">
                <hy:params>
                    <hy:param name="xpath_expression" type="com.hyfinity.xpath.XPath">
                        <xsl:call-template name="output-location-xpath">
                            <xsl:with-param name="locationXPath" select="$datasource/@location"/>
                        </xsl:call-template>
                        <xsl:if test="$datasource/@type = 'db' and $datasource/@use = 'editMany'" >
                            <xsl:text>/record</xsl:text>
                        </xsl:if>
                    </hy:param>
                </hy:params>
            </hy:target>
        </hy:clause>

    </xsl:template>


    <xsl:template name="output-location-xpath">
        <xsl:param name="locationXPath"/>

        <!-- Need to check namespaces used in this xpath -->
        <xsl:choose>
            <xsl:when test="contains($locationXPath, '&lt;*')">
                <xsl:value-of select="substring-before($locationXPath, '&lt;*')"/>

                <xsl:variable name="nsStart" select="substring-after($locationXPath, '&lt;*')"/>

                <xsl:value-of select="$rulebaseNamespacesSet/namespaces/namespace[@value = substring-before($nsStart, '*>')]/@prefix"/>

                <xsl:call-template name="output-location-xpath">
                    <xsl:with-param name="locationXPath" select="substring-after($nsStart, '*>')"/>
                </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
                <xsl:value-of select="$locationXPath"/>
            </xsl:otherwise>
        </xsl:choose>

    </xsl:template>




    <!-- looks through the current rulebase, and any new datasource details
         to find the complete list of all the namespaces that will need to be defined
         on the resultant logicsheet -->
    <xsl:template name="find-namespaces">
        <namespaces xmlns="">
            <!-- first find all the existing anmesapce definitions -->
            <xsl:for-each select="$rulebase/namespace::*[local-name(.) != '' and local-name(.) != 'xml']">
                <namespace prefix="{local-name()}" value="{.}"/>
            </xsl:for-each>

            <!-- now check through each location xpath -->
            <xsl:variable name="locationNamespaces">
                <namespaces xmlns="">
                    <xsl:for-each select="$incomingActions/fm:data_sources/fm:data_source[not(@generated = 'true') and @location != '']" >
                        <xsl:call-template name="find-location-namespaces">
                            <xsl:with-param name="location" select="@location"/>
                        </xsl:call-template>
                    </xsl:for-each>
                </namespaces>
            </xsl:variable>

            <xsl:variable name="locationNamespacesSet" select="exsl:node-set($locationNamespaces)"/>

            <xsl:call-template name="process-location-namespaces" >
                <xsl:with-param name="ns" select="$locationNamespacesSet/namespaces/namespace[1]"/>
            </xsl:call-template>
        </namespaces>
    </xsl:template>


    <xsl:template name="find-location-namespaces">
        <xsl:param name="location"/>

        <xsl:if test="contains($location, '&lt;*')">
            <xsl:variable name="nsStart" select="substring-after($location, '&lt;*')"/>
            <xsl:variable name="ns" select="substring-before($nsStart, '*>')"/>
            <!-- Check that this namespace doesnt already exist on the rulebase -->
            <xsl:if test="not($rulebase/namespace::*[. = $ns])" >
                <namespace>
                    <xsl:value-of select="$ns"/>
                </namespace>
            </xsl:if>
            <xsl:call-template name="find-location-namespaces">
                <xsl:with-param name="location" select="substring-after($nsStart, '*>')"/>
            </xsl:call-template>
        </xsl:if>
    </xsl:template>


    <xsl:template name="process-location-namespaces">
        <xsl:param name="ns"/>
        <xsl:param name="usedPrefixes" />

        <xsl:if test="$ns != ''" >

            <!-- check if this namespace is a duplicate that we have already processed. -->
            <xsl:choose>
                <xsl:when test="$ns[(. = preceding-sibling::namespace)]">
                    <!-- duplciate, so just move on to process the next one -->
                    <xsl:call-template name="process-location-namespaces" >
                        <xsl:with-param name="ns" select="$ns/following-sibling::*[1]"/>
                        <xsl:with-param name="usedPrefixes" select="$usedPrefixes"/>
                    </xsl:call-template>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:variable name="newPrefix">
                        <xsl:call-template name="findNextPrefix">
                            <xsl:with-param name="start">ns</xsl:with-param>
                            <xsl:with-param name="count">1</xsl:with-param>
                            <xsl:with-param name="usedPrefixes" select="$usedPrefixes"/>
                        </xsl:call-template>
                    </xsl:variable>

                    <namespace value="{$ns}" prefix="{$newPrefix}">
                    </namespace>

                    <xsl:call-template name="process-location-namespaces" >
                        <xsl:with-param name="ns" select="$ns/following-sibling::*[1]"/>
                        <xsl:with-param name="usedPrefixes" select="concat($usedPrefixes, ':', $newPrefix)"/>
                    </xsl:call-template>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:if>
    </xsl:template>


    <xsl:template name="findNextPrefix">
        <xsl:param name="start">ns</xsl:param>
        <xsl:param name="count">1</xsl:param>
        <xsl:param name="usedPrefixes" />

        <xsl:choose>
            <!-- check if prefix already exists on the rulebase -->
            <xsl:when test="$rulebase/namespace::*[local-name() = concat($start, $count)]">
                <xsl:call-template name="findNextPrefix">
                    <xsl:with-param name="start" select="$start"/>
                    <xsl:with-param name="count" select="$count + 1"/>
                    <xsl:with-param name="usedPrefixes" select="$usedPrefixes"/>
                </xsl:call-template>
            </xsl:when>
            <!-- check if prefix has already been used for a new namespace -->
            <xsl:when test="contains(concat($usedPrefixes, ':'), concat(':', $start, $count, ':'))">
                <xsl:call-template name="findNextPrefix">
                    <xsl:with-param name="start" select="$start"/>
                    <xsl:with-param name="count" select="$count + 1"/>
                    <xsl:with-param name="usedPrefixes" select="$usedPrefixes"/>
                </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
                <xsl:value-of select="$start"/>
                <xsl:value-of select="$count"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>

</xsl:stylesheet>
